@@ -40,39 +40,35 @@
 
 /*
 **
-** mm_core.c -- Low-Level Memory Allocation
+**  mm_core.c -- Low-level Shared Memory API
 **
 */
 
 #define MM_PRIVATE
 #include "mm.h"
 
-#if !defined(MM_NA)
-
 /*
  * Some global variables
  */
 
 #if defined(MM_SEMT_FCNTL)
 /* lock/unlock structures for fcntl() */
-static struct flock mm_core_dolock;
+static struct flock mm_core_dolock_rd;
+static struct flock mm_core_dolock_rw;
 static struct flock mm_core_dounlock;
 #endif
 
 #if defined(MM_SEMT_IPCSEM)
 /* lock/unlock structures for semop() */
-static struct sembuf mm_core_dolock[] = {
-    { 0, 0, 0 },        /* wait for semaphore */
-    { 0, 1, SEM_UNDO }  /* increment semaphore */
-};
-static struct sembuf mm_core_dounlock[] = {
-    { 0, -1, SEM_UNDO } /* derements semaphore */
-};
+static union  semun  mm_core_semctlarg;
+static struct sembuf mm_core_dolock[2];
+static struct sembuf mm_core_dounlock[1];
 #endif
 
-#if defined(MM_SHMT_MMZERO) || defined(MM_SHMT_MMFILE)
-/* mapping offset for mmap() */
-static int mm_core_mmap_offset = 0;
+#if defined(MM_SHMT_MMFILE) || defined(MM_SHMT_MMPOSX)
+static size_t mm_core_mapoffset = 0;           /* we use own file */
+#elif defined(MM_SHMT_MMZERO)
+static size_t mm_core_mapoffset = 1024*1024*1; /* we share with other apps */
 #endif
 
 static void mm_core_init(void)
@@ -80,16 +76,32 @@ static void mm_core_init(void)
     static int initialized = FALSE;
     if (!initialized) {
 #if defined(MM_SEMT_FCNTL)
-        mm_core_dolock.l_whence   = SEEK_SET; /* from current point */
-        mm_core_dolock.l_start    = 0;        /* -"- */
-        mm_core_dolock.l_len      = 0;        /* until end of file */
-        mm_core_dolock.l_type     = F_WRLCK;  /* set exclusive/write lock */
-        mm_core_dolock.l_pid      = 0;        /* pid not actually interesting */
-        mm_core_dounlock.l_whence = SEEK_SET; /* from current point */
-        mm_core_dounlock.l_start  = 0;        /* -"- */
-        mm_core_dounlock.l_len    = 0;        /* until end of file */
-        mm_core_dounlock.l_type   = F_UNLCK;  /* unlock */
-        mm_core_dounlock.l_pid    = 0;        /* pid not actually interesting */
+        mm_core_dolock_rd.l_whence   = SEEK_SET; /* from current point */
+        mm_core_dolock_rd.l_start    = 0;        /* -"- */
+        mm_core_dolock_rd.l_len      = 0;        /* until end of file */
+        mm_core_dolock_rd.l_type     = F_RDLCK;  /* set shard/read lock */
+        mm_core_dolock_rd.l_pid      = 0;        /* pid not actually interesting */
+        mm_core_dolock_rw.l_whence   = SEEK_SET; /* from current point */
+        mm_core_dolock_rw.l_start    = 0;        /* -"- */
+        mm_core_dolock_rw.l_len      = 0;        /* until end of file */
+        mm_core_dolock_rw.l_type     = F_WRLCK;  /* set exclusive/read-write lock */
+        mm_core_dolock_rw.l_pid      = 0;        /* pid not actually interesting */
+        mm_core_dounlock.l_whence    = SEEK_SET; /* from current point */
+        mm_core_dounlock.l_start     = 0;        /* -"- */
+        mm_core_dounlock.l_len       = 0;        /* until end of file */
+        mm_core_dounlock.l_type      = F_UNLCK;  /* unlock */
+        mm_core_dounlock.l_pid       = 0;        /* pid not actually interesting */
+#endif
+#if defined(MM_SEMT_IPCSEM)
+        mm_core_dolock[0].sem_num   = 0;
+        mm_core_dolock[0].sem_op    = 0;
+        mm_core_dolock[0].sem_flg   = 0;
+        mm_core_dolock[1].sem_num   = 0;
+        mm_core_dolock[1].sem_op    = 1;
+        mm_core_dolock[1].sem_flg   = SEM_UNDO;
+        mm_core_dounlock[0].sem_num = 0;
+        mm_core_dounlock[0].sem_op  = -1;
+        mm_core_dounlock[0].sem_flg = SEM_UNDO;
 #endif
         initialized = TRUE;
     }
@@ -138,7 +150,7 @@ static size_t mm_core_pagesize(void)
 #elif defined(MM_VMPS_SYSCONF)
         pagesize = sysconf(_SC_PAGESIZE);
 #else
-        pagesize = MM_DEFAULT_PAGESIZE;
+        pagesize = MM_CORE_DEFAULT_PAGESIZE;
 #endif
     return pagesize;
 }
@@ -158,16 +170,24 @@ size_t mm_core_align2word(size_t size)
     return ((1+((size-1) / SIZEOF_mem_word)) * SIZEOF_mem_word);
 }
 
+size_t mm_core_maxsegsize(void)
+{
+    return MM_SHM_MAXSEGSIZE;
+}
+
 /*
  * Create a shared memory area
  */
 
-void *mm_core_create(size_t usersize, char *file)
+void *mm_core_create(size_t usersize, const char *file)
 {
     mem_core *mc;
     void *area = ((void *)-1);
     int fdmem = 0;
     int fdsem = 0;
+#if defined(MM_SEMT_IPCSEM)
+    int fdsem_rd = 0;
+#endif
 #if defined(MM_SHMT_MMPOSX) || defined(MM_SHMT_MMFILE)
     char *fnmem;
 #endif
@@ -175,7 +195,7 @@ void *mm_core_create(size_t usersize, char *file)
     char *fnsem;
 #endif
     size_t size;
-#if defined(MM_SHMT_MMZERO) || defined(MM_SHMT_MMPOSIX) || defined(MM_SHMT_MMFILE)
+#if defined(MM_SHMT_MMZERO) || defined(MM_SHMT_MMPOSX) || defined(MM_SHMT_MMFILE)
     int zero = 0;
 #endif
 #if defined(MM_SHMT_IPCSHM)
@@ -190,10 +210,10 @@ void *mm_core_create(size_t usersize, char *file)
     char filename[MM_MAXPATH];
 
     mm_core_init();
-    if (usersize == 0)
-        usersize = MM_DEFAULT_SIZE;
+    if (usersize <= 0 || usersize >= mm_core_maxsegsize())
+        usersize = mm_core_maxsegsize();
     if (file == NULL) {
-        sprintf(filename, MM_DEFAULT_FILE, (int)getpid());
+        sprintf(filename, MM_CORE_DEFAULT_FILE, (int)getpid());
         file = filename;
     }
     size = mm_core_align2page(usersize+SIZEOF_mem_core);
@@ -209,81 +229,91 @@ void *mm_core_create(size_t usersize, char *file)
 
 #if defined(MM_SHMT_MMANON)
     if ((area = (void *)mmap(NULL, size, PROT_READ|PROT_WRITE, 
-                             MAP_ANON|MAP_SHARED, -1, 0)) == ((void *)-1))
-        FAIL;
+                             MAP_ANON|MAP_SHARED, -1, 0)) == MAP_FAILED)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to memory map anonymous area");
 #endif /* MM_SHMT_MMANON */
 
 #if defined(MM_SHMT_MMPOSX)
     shm_unlink(fnmem); /* Ok when it fails */
-    if ((fdmem = shm_open(fnmem, O_RDWR, MM_CORE_FILEMODE)) == -1)
-        FAIL;
-    if (ftruncate(fdmem, sizeof(mem_core)) == -1)
-        FAIL;
+    if ((fdmem = shm_open(fnmem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open tempfile");
+    if (ftruncate(fdmem, mm_core_mapoffset+size) == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to truncate tempfile");
     write(fdmem, &zero, sizeof(zero));
     if ((area = (void *)mmap(NULL, size, PROT_READ|PROT_WRITE, 
-                             MAP_SHARED, fdmem, 0)) == ((void *)-1))
-        FAIL;
+                             MAP_SHARED, fdmem, mm_core_mapoffset)) == MAP_FAILED)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to memory map tempfile");
     shm_unlink(fnmem);
+    mm_core_mapoffset += size;
 #endif /* MM_SHMT_MMPOSX */
 
 #if defined(MM_SHMT_MMZERO)
     if ((fdmem = open("/dev/zero", O_RDWR, MM_CORE_FILEMODE)) == -1)
-        FAIL;
-    if (ftruncate(fdmem, mm_core_mmap_offset+sizeof(mem_core)) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open /dev/zero");
+    if (lseek(fdmem, mm_core_mapoffset+size, SEEK_SET) == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to seek in /dev/zero");
     write(fdmem, &zero, sizeof(zero));
     if ((area = (void *)mmap(NULL, size, PROT_READ|PROT_WRITE, 
-                             MAP_SHARED, fdmem, mm_core_mmap_offset)) == ((void *)-1))
-        FAIL;
-    mm_core_mmap_offset += size;
+                             MAP_SHARED, fdmem, mm_core_mapoffset)) == MAP_FAILED)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to memory map /dev/zero");
+    mm_core_mapoffset += size;
 #endif /* MM_SHMT_MMZERO */
 
 #if defined(MM_SHMT_MMFILE)
     unlink(fnmem);
     if ((fdmem = open(fnmem, O_RDWR|O_CREAT, MM_CORE_FILEMODE)) == -1)
-        FAIL;
-    if (ftruncate(fdmem, mm_core_mmap_offset+sizeof(mem_core)) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open memory file");
+    if (ftruncate(fdmem, mm_core_mapoffset+size) == -1)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to truncate memory file");
     write(fdmem, &zero, sizeof(zero));
     if ((area = (void *)mmap(NULL, size, PROT_READ|PROT_WRITE, 
-                             MAP_SHARED, fdmem, mm_core_mmap_offset)) == ((void *)-1))
-        FAIL;
-    mm_core_mmap_offset += size;
+                             MAP_SHARED, fdmem, mm_core_mapoffset)) == MAP_FAILED)
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to memory map memory file");
+    mm_core_mapoffset += size;
 #endif /* MM_SHMT_MMFILE */
 
 #if defined(MM_SHMT_IPCSHM)
     if ((fdmem = shmget(IPC_PRIVATE, size, (SHM_R|SHM_W|IPC_CREAT))) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire shared memory segment");
     if ((area = (void *)shmat(fdmem, NULL, 0)) == ((void *)-1))
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to attach shared memory");
     if (shmctl(fdmem, IPC_STAT, &shmbuf) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to get status of shared memory");
     shmbuf.shm_perm.uid = getuid();
     shmbuf.shm_perm.gid = getgid();
     if (shmctl(fdmem, IPC_SET, &shmbuf) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to set status of shared memory");
     if (shmctl(fdmem, IPC_RMID, NULL) == -1)
-        FAIL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to remove shared memory in advance");
 #endif /* MM_SHMT_IPCSHM */
 
 #if defined(MM_SEMT_FLOCK)
     unlink(fnsem);
     if ((fdsem = open(fnsem, O_WRONLY|O_CREAT, MM_CORE_FILEMODE)) == -1)
-        return NULL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open semaphore file");
 #endif /* MM_SEMT_FLOCK */
 
 #if defined(MM_SEMT_FCNTL)
     unlink(fnsem);
     if ((fdsem = open(fnsem, O_WRONLY|O_CREAT, MM_CORE_FILEMODE)) == -1)
-        return NULL;
+        FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to open semaphore file");
 #endif /* MM_SEMT_FCNTL */
 
 #if defined(MM_SEMT_IPCSEM)
-    fdsem = semget(IPC_PRIVATE, 1, (IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR));
+    fdsem = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
     if (fdsem == -1 && errno == EEXIST)
-        fdsem = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
+        fdsem = semget(IPC_PRIVATE, 1, IPC_EXCL|S_IRUSR|S_IWUSR);
         if (fdsem == -1)
-            FAIL;
+            FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
+    mm_core_semctlarg.val = 0;
+    semctl(fdsem, 0, SETVAL, mm_core_semctlarg);
+    fdsem_rd = semget(IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL|S_IRUSR|S_IWUSR);
+    if (fdsem_rd == -1 && errno == EEXIST)
+        fdsem_rd = semget(IPC_PRIVATE, 1, IPC_EXCL|S_IRUSR|S_IWUSR);
+        if (fdsem_rd == -1)
+            FAIL(MM_ERR_CORE|MM_ERR_SYSTEM, "failed to acquire semaphore");
+    mm_core_semctlarg.val = 0;
+    semctl(fdsem_rd, 0, SETVAL, mm_core_semctlarg);
 #endif /* MM_SEMT_IPCSEM */
 
     /*
@@ -302,6 +332,10 @@ void *mm_core_create(size_t usersize, char *file)
 #else
     mc->mc_fdsem    = fdsem;
 #endif
+#if defined(MM_SEMT_IPCSEM)
+    mc->mc_fdsem_rd = fdsem_rd;
+    mc->mc_readers  = 0;
+#endif
 #if defined(MM_SHMT_MMFILE)
     memcpy(mc->mc_fnmem, fnmem, MM_MAXPATH);
 #endif
@@ -318,7 +352,7 @@ void *mm_core_create(size_t usersize, char *file)
      * clean-up sequence (CUS) for error situation
      */
     BEGIN_FAILURE
-#if defined(MM_SHMT_MMANON) || defined(MM_SHMT_MMZERO) || defined(MM_SHMT_MMPOSIX) || defined(MM_SHMT_MMFILE)
+#if defined(MM_SHMT_MMANON) || defined(MM_SHMT_MMZERO) || defined(MM_SHMT_MMPOSX) || defined(MM_SHMT_MMFILE)
     if (area != ((void *)-1))
         munmap((caddr_t)area, size);
 #endif
@@ -341,6 +375,8 @@ void *mm_core_create(size_t usersize, char *file)
 #if defined(MM_SEMT_IPCSEM)
     if (fdsem != -1)
         semctl(fdsem, 0, IPC_RMID, 0);
+    if (fdsem_rd != -1)
+        semctl(fdsem_rd, 0, IPC_RMID, 0);
 #endif
 #if defined(MM_SHMT_MMFILE)
     unlink(fnmem);
@@ -357,6 +393,9 @@ void mm_core_delete(void *core)
     mem_core *mc;
     int fdmem;
     int fdsem;
+#if defined(MM_SEMT_IPCSEM)
+    int fdsem_rd;
+#endif
     size_t size;
 #if defined(MM_SHMT_MMFILE)
     char fnmem[MM_MAXPATH];
@@ -373,6 +412,9 @@ void mm_core_delete(void *core)
 #if !defined(MM_SEMT_FLOCK)
     fdsem = mc->mc_fdsem;
 #endif
+#if defined(MM_SEMT_IPCSEM)
+    fdsem_rd = mc->mc_fdsem_rd;
+#endif
 #if defined(MM_SEMT_FLOCK)
     fdsem = mm_core_getfdsem(mc);
 #endif
@@ -404,11 +446,12 @@ void mm_core_delete(void *core)
 #endif
 #if defined(MM_SEMT_IPCSEM)
     semctl(fdsem, 0, IPC_RMID, 0);
+    semctl(fdsem_rd, 0, IPC_RMID, 0);
 #endif
     return;
 }
 
-size_t mm_core_size(void *core)
+size_t mm_core_size(const void *core)
 {
     mem_core *mc;
 
@@ -418,7 +461,7 @@ size_t mm_core_size(void *core)
     return (mc->mc_usize);
 }
 
-int mm_core_lock(void *core)
+int mm_core_lock(const void *core, mm_lock_mode mode)
 {
     mem_core *mc;
     int rc;
@@ -435,25 +478,40 @@ int mm_core_lock(void *core)
 #endif
 
 #if defined(MM_SEMT_FCNTL)
-    errno = 0;
-    while (((rc = fcntl(fdsem, F_SETLKW, &mm_core_dolock)) < 0) && (errno == EINTR)) ;
+    if (mode == MM_LOCK_RD)
+        while (((rc = fcntl(fdsem, F_SETLKW, &mm_core_dolock_rd)) < 0) && (errno == EINTR)) ;
+    else
+        while (((rc = fcntl(fdsem, F_SETLKW, &mm_core_dolock_rw)) < 0) && (errno == EINTR)) ;
 #endif
 #if defined(MM_SEMT_FLOCK)
-    errno = 0;
-    while (((rc = flock(fdsem, LOCK_EX)) < 0) && (errno == EINTR)) ;
+    if (mode == MM_LOCK_RD)
+        while (((rc = flock(fdsem, LOCK_SH)) < 0) && (errno == EINTR)) ;
+    else
+        while (((rc = flock(fdsem, LOCK_EX)) < 0) && (errno == EINTR)) ;
 #endif
 #if defined(MM_SEMT_IPCSEM)
-    errno = 0;
-    while (((rc = semop(fdsem, mm_core_dolock, 2) < 0) && (errno == EINTR)) ;
+    if (mode == MM_LOCK_RD) {
+        while (((rc = semop(mc->mc_fdsem_rd, mm_core_dolock, 2)) < 0) && (errno == EINTR)) ;
+        mc->mc_readers++;
+        if (mc->mc_readers == 1)
+            while (((rc = semop(fdsem, mm_core_dolock, 2)) < 0) && (errno == EINTR)) ;
+        while (((rc = semop(mc->mc_fdsem_rd, mm_core_dounlock, 1)) < 0) && (errno == EINTR)) ;
+    }
+    else {
+        while (((rc = semop(fdsem, mm_core_dolock, 2)) < 0) && (errno == EINTR)) ;
+    }
+    mc->mc_lockmode = mode;
 #endif
-    if (rc < 0)
+    if (rc < 0) {
+        ERR(MM_ERR_CORE|MM_ERR_SYSTEM, "Failed to lock");
         rc = FALSE;
+    }
     else
         rc = TRUE;
     return rc;
 }
 
-int mm_core_unlock(void *core)
+int mm_core_unlock(const void *core)
 {
     mem_core *mc;
     int rc;
@@ -470,23 +528,30 @@ int mm_core_unlock(void *core)
 #endif
 
 #if defined(MM_SEMT_FCNTL)
-    errno = 0;
     while (((rc = fcntl(fdsem, F_SETLKW, &mm_core_dounlock)) < 0) && (errno == EINTR)) ;
 #endif
 #if defined(MM_SEMT_FLOCK)
-    errno = 0;
     while (((rc = flock(fdsem, LOCK_UN)) < 0) && (errno == EINTR)) ;
 #endif
 #if defined(MM_SEMT_IPCSEM)
-    errno = 0;
-    while (((rc = semop(fdsem, mm_core_dounlock, 1) < 0) && (errno == EINTR)) ;
+    if (mc->mc_lockmode == MM_LOCK_RD) {
+        while (((rc = semop(mc->mc_fdsem_rd, mm_core_dolock, 2)) < 0) && (errno == EINTR)) ;
+        mc->mc_readers--;
+        if (mc->mc_readers == 0)
+            while (((rc = semop(fdsem, mm_core_dounlock, 1)) < 0) && (errno == EINTR)) ;
+        while (((rc = semop(mc->mc_fdsem_rd, mm_core_dounlock, 1)) < 0) && (errno == EINTR)) ;
+    }
+    else {
+        while (((rc = semop(fdsem, mm_core_dounlock, 1)) < 0) && (errno == EINTR)) ;
+    }
 #endif
-    if (rc < 0)
+    if (rc < 0) {
+        ERR(MM_ERR_CORE|MM_ERR_SYSTEM, "Failed to unlock");
         rc = FALSE;
+    }
     else
         rc = TRUE;
     return rc;
 }
 
-#endif /* MM_NA */
 /*EOF*/
